from __future__ import annotations
from abc import ABC
import astropy.units as un
from astropy.units import Unit
from astropy.units import Quantity
import numpy as np
from ..units import check_units
from ..units import beam_equivalencies
from .parameter import Parameter
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..base import Dataset


class Image(Parameter, ABC):
    def __init__(self, **kwargs):
        """
        Class that represents an image. Inherits from parameter.
        :param kwargs: Parameter arguments
        """
        super().__init__(**kwargs)

        if self.data is not None:
            if self.data.ndim == 2:
                self.imsize = [self.data.shape[0], self.data.shape[1]]
            elif self.data.ndim == 3:
                self.imsize = [self.data.shape[1], self.data.shape[2]]
            else:
                raise ValueError("The image dimensions should be greater than 1")

        if check_units(self.cellsize, un.rad):
            self.cellsize = self.cellsize.to(un.rad)
        else:
            raise ValueError("Units for cellsize should be in units of Angle")

    def transform_intensity_units(self, intensity: Quantity = None, beam_area: Quantity = None, unit: Unit = None):
        if intensity is not None and beam_area is not None and unit is not None:
            if beam_area.unit == self.pixel_area.unit:
                beam_equiv = beam_equivalencies(beam_area=beam_area / self.pixel_area, unit=unit)
            else:
                raise ValueError("Beam and image areas need to have same units")
            return intensity.to((un.Jy / unit), equivalencies=beam_equiv)
        else:
            raise ValueError("Intensity, beam_area and unit need to be different from None")

    def calculate_noise(self, measurements: Dataset = None) -> None:
        if measurements is None:
            pass
        else:
            if measurements.feed_kind is None:
                measurements.check_feed()

            if measurements.parhands_weight_sum > 0.0:
                parhands_noise = 0.5 * np.sqrt(1.0 / measurements.parhands_weight_sum)
            else:
                parhands_noise = 0.0

            if measurements.crossedhands_weight_sum > 0.0:
                crosshands_noise = 0.5 * np.sqrt(1.0 / measurements.crossedhands_weight_sum)
            else:
                crosshands_noise = 0.0

            if measurements.feed_kind == "linear":
                if self.name == "I" or self.name == "Q":
                    self.noise = parhands_noise * (un.Jy / un.beam)
                elif self.name == "U" or self.name == "V":
                    self.noise = crosshands_noise * (un.Jy / un.beam)

            elif measurements.feed_kind == "circular":
                if self.name == "I" or self.name == "V":
                    self.noise = parhands_noise * (un.Jy / un.beam)
                elif self.name == "Q" or self.name == "U":
                    self.noise = crosshands_noise * (un.Jy / un.beam)



import logging
import xarray as xr

pol_dict = {
    0: None,
    1: "I_s",
    2: "Q_s",
    3: "U_s",
    4: "V_s",
    5: "RR",
    6: "RL",
    7: "LR",
    8: "LL",
    9: "XX",
    10: "XY",
    11: "YX",
    12: "YY",
    13: "RX",
    14: "RY",
    15: "LX",
    16: "LY",
    17: "XR",
    18: "XL",
    19: "YR",
    20: "YL",
    21: "PP",
    22: "PQ",
    23: "QP",
    24: "QQ",
    25: "RCircular",
    26: "LCircular",
    27: "Linear",
    28: "Ptotal",
    29: "Plinear",
    30: "PFtotal",
    31: "PFlinear",
    32: "Pangle"
}


class Polarization:
    def __init__(self, _id: int = None, dataset: xr.Dataset = None):
        """
        Class that represents the polarization table of the partitioned Measurement Set

        Parameters
        ----------
        _id : int
              ID of the object
        dataset : xarray.Dataset
                  xarray Dataset with the polarization table
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self._id = _id
        self.dataset = dataset

        self.corrs = []
        self.ncorrs = 0
        self.feed_kind = None
        if dataset is not None and _id is not None:
            self.corrs = dataset.CORR_TYPE.data.squeeze().compute()
            corr_string_list = [pol_dict[corr] for corr in self.corrs] # TODO: self.corr_names can be a string list -
            # no need to be a dictionary
            self.corrs_names = {i: corr_string_list[i] for i in range(0, len(corr_string_list))}
            linear_feed = ["XX", "XY", "YX", "YY"]
            if any(i in linear_feed for i in [*self.corrs_names.values()]):
                self.feed_kind = "linear"
            else:
                self.feed_kind = "circular"
            self.ncorrs = dataset.NUM_CORR.data.squeeze().compute()



from abc import ABC
import dask.array as da
import numpy as np
from ..wscheme import WeightingScheme


class Robust(WeightingScheme, ABC):
    def __init__(self, robust_parameter: float = None, **kwargs):
        """
        :param robust_parameter: Briggs/Robust parameter
        :param kwargs: WeightingScheme arguments
        """
        super().__init__(**kwargs)
        self.robust_parameter = robust_parameter

    @property
    def robust_parameter(self):
        return self.__robust_parameter

    @robust_parameter.setter
    def robust_parameter(self, val):
        if -2.0 <= val <= 2.0:
            self.__robust_parameter = val
        else:
            raise ValueError("Robust parameter needs to be between [-2.0, 2.0]")

    def calculate_f2(self, sum_gridweights_squared, sum_weights, robust_parameter: float = None):
        """
        :param sum_gridweights_squared: Sum of the gridded weights squared
        :param sum_weights: Sum of the non-gridded weights
        :param robust_parameter: Robust/Briggs parameter
        :return: The f^2 term in the Briggs equation
        """
        if robust_parameter is None:
            robust_parameter = -1.0 * self.robust_parameter
        num = (5. * 10 ** robust_parameter) ** 2
        den = sum_gridweights_squared / sum_weights
        if den.any() == 0.0:
            raise ZeroDivisionError
        else:
            return num / den

    def transform(self) -> None:
        """
        This function calculates the robust weights and transforms them according to this scheme,
        """
        if self.gridder is not None:
            dataset = self.input_data

            for ms in dataset.ms_list:
                weight = ms.visibilities.weight
                idx, bin_count, w_k = self.gridder.grid_weights(ms)
                original_weight_sum = da.apply_along_axis(np.sum, 0, weight.data)
                unique_idx = da.unique(idx).compute()
                gridded_weights = bin_count[unique_idx]  # Gridded weight for each u,v coordinate on the grid
                gridded_weightsquared_sum = da.apply_along_axis(np.sum, 0, gridded_weights * gridded_weights)
                f2 = self.calculate_f2(gridded_weightsquared_sum, original_weight_sum)
                weight /= (1.0 + f2 * w_k)
        else:
            raise TypeError("gridding attribute cannot be None when calculting robust weights")



from abc import ABC
import logging
from .io import Io
from astropy.io import fits
import dask.array as da
import xarray as xr
from astropy.io.fits import Header
import astropy.units as un
import numpy as np
from typing import Union


class FITS(Io, ABC):
    def __init__(self, hdu: int = 0, memmap: bool = True, lazy_load_hdus: bool = False, use_dask: bool = True,
                 chunks: Union[int, tuple, str] = 'auto', **kwargs):
        """
        I/O class to handle FITS files.
        :param hdu: HDU of the FITS file
        :param memmap: Whether to use memory map or to load the whole image in memory
        :param lazy_load_hdus: Whether to load HDUs on a lazy manner
        :param use_dask: Whether to use dask arrays or not.
        :param chunks: How to chunk the array. Must be one of the following forms:
                - A blocksize like 1000.
                - A blockshape like (1000, 1000).
                - Explicit sizes of all blocks along all dimensions like
                ((1000, 1000, 500), (400, 400)).
                - A size in bytes, like "100 MiB" which will choose a uniform
                block-like shape
                - The word "auto" which acts like the above, but uses a configuration
                value ``array.chunk-size`` for the chunk size
                -1 or None as a blocksize indicate the size of the corresponding
                dimension.
        :param kwargs: IO arguments
        """
        super().__init__(**kwargs)
        self.hdu = hdu
        self.memmap = memmap
        self.lazy_load_hdus = lazy_load_hdus
        self.header = None
        self.use_dask = use_dask
        self.chunks = chunks
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

    def get_coordinates(self, hdul) -> dict:
        """
        Function that gets coordinates from header
        :param hdul: HDU List
        :return: Dictionary that represents the xarray coordinates
        """
        ndims = hdul[self.hdu].data.ndim
        m = self.header["NAXIS1"]
        n = self.header["NAXIS2"]
        if ndims == 3:
            z = self.header["NAXIS3"]
            cdelt3 = self.header["CDELT3"] * un.Hz
            crval3 = self.header["CRVAL3"] * un.Hz

        crpix1 = self.header["CRPIX1"]
        crpix2 = self.header["CRPIX2"]
        cdelt1 = self.header["CDELT1"] * un.deg
        cdelt2 = self.header["CDELT2"] * un.deg

        i_index_dask = da.arange(0, n, dtype=np.float32)
        j_index_dask = da.arange(0, m, dtype=np.float32)
        j_grid, i_grid = da.meshgrid(i_index_dask, j_index_dask)

        z_index = da.arange(0, z, dtype=np.float32)

        ra_dask = (j_index_dask - crpix1) * cdelt1
        dec_dask = (i_index_dask - crpix2) * cdelt2
        ra_grid, dec_grid = da.meshgrid(ra_dask, dec_dask)
        if ndims == 3:
            freqs = (z_index * cdelt3.value + crval3.value) * un.Hz
            coords = dict(
                Z=(["nu"], z_index.astype(np.int32)),
                FREQS=(["nu"], freqs),
                X=(["x", "y"], j_grid.astype(np.int32)),
                Y=(["x", "y"], i_grid.astype(np.int32)),
                RA=(["x", "y"], ra_grid),
                DEC=(["x", "y"], dec_grid),
            )

        elif ndims == 2:
            coords = dict(
                X=(["x", "y"], j_grid.astype(np.int32)),
                Y=(["x", "y"], i_grid.astype(np.int32)),
                RA=(["x", "y"], ra_grid),
                DEC=(["x", "y"], dec_grid),
            )

        return coords

    def read(self):
        with fits.open(self.input_name, memmap=self.memmap, lazy_load_hdus=self.lazy_load_hdus) as hdul:
            self.header = hdul[self.hdu].header
            ndims = hdul[self.hdu].data.squeeze().ndim
            coords = self.get_coordinates(hdul)
            if ndims == 3:
                dims = ["nu", "x", "y"]
            elif ndims == 2:
                dims = ["x", "y"]
            if self.use_dask:
                data = da.from_array(hdul[self.hdu].data, chunks=self.chunks).squeeze()
            else:
                data = hdul[self.hdu].data.squeeze()
            return xr.DataArray(data, dims=dims, coords=coords,
                                attrs=dict(hdul[self.hdu].header))

    def write(self, fitsdata: xr.DataArray, output_name: str = None, overwrite: bool = True):
        data = fitsdata.data.compute()

        if self.header is None:
            attrs = fitsdata.attrs
            if attrs:
                header = Header(attrs)
            else:
                header = None
        else:
            header = self.header

        if output_name is None:
            fits.writeto(filename=self.output_name, data=data, header=header, overwrite=overwrite)
        else:
            fits.writeto(filename=output_name, data=data, header=header, overwrite=overwrite)



import xarray as xr
import numpy as np
from .antenna import Antenna
import logging
from itertools import combinations
import xarray.ufuncs as xarrfunc
import dask.array as da
from typing import Union
import astropy.units as u


class Baseline:
    def __init__(self, antenna: Antenna = None, dask_idx: bool = False):
        """
        Class that represents the relationship between two antennas.
        The resulting baseline dataset has the following columns:
        - BASELINE_LENGTH
        - ANTENNA1
        - ANTENNA2

        Parameters
        ----------
        antenna : Antenna
                  Antenna object.
        dask_idx : bool
                   Boolean to use or not use indexes as dask arrays.
        """

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self.dataset = None
        self.max_baseline = 0.0 * u.m
        self.min_baseline = 0.0 * u.m

        if antenna is not None:
            ids = antenna.dataset.ROWID.data.compute()
            combs = np.array(list(combinations(ids, 2)))

            antenna1 = antenna.dataset.sel(row=combs[:, 0])
            antenna2 = antenna.dataset.sel(row=combs[:, 1])

            baseline = antenna1.POSITION - antenna2.POSITION
            baseline_length = xarrfunc.sqrt(
                xarrfunc.square(baseline[:, 0]) + xarrfunc.square(baseline[:, 1]) + xarrfunc.square(baseline[:, 2]))
            baseline_length = baseline_length.data.astype(np.float32) * u.m

            row_id = np.arange(len(combs[:, 0]))
            ant1_id = combs[:, 0]
            ant2_id = combs[:, 1]
            if dask_idx:
                row_id = da.from_array(row_id)
                ant1_id = da.from_array(ant1_id)
                ant2_id = da.from_array(ant2_id)

            ds = xr.Dataset(
                data_vars=dict(
                    BASELINE_LENGTH=(["row"], baseline_length),
                    ANTENNA1=(["row"], ant1_id.astype(np.int32)),
                    ANTENNA2=(["row"], ant2_id.astype(np.int32)),

                ),
                coords=dict(
                    ROWID=(["row"], row_id.astype(np.int32)),
                ),
                attrs=dict(description="Baseline-Antenna relationship"))
            self.dataset = ds
            self.max_baseline = self.dataset.BASELINE_LENGTH.max().data.compute()
            self.min_baseline = self.dataset.BASELINE_LENGTH.min().data.compute()

    def filter_by_antennas(self, antenna1: Union[da.Array, xr.DataArray, np.array, int],
                           antenna2: Union[da.Array, xr.DataArray, np.array, int], drop: bool = True) -> xr.Dataset:
        """
        :param antenna1: Array or scalar of ids referencing ANTENNA1
        :param antenna2: Array or scalar of ids referencing ANTENNA2
        :param drop: Drop nan values when doing where on the dataset
        :return: Filtered xarray dataset
        """
        if self.dataset is not None:
            filtered_dataset = self.dataset.where((self.dataset.ANTENNA1.isin(antenna1)) &
                                                  (self.dataset.ANTENNA2.isin(antenna2)), drop=drop)
            return filtered_dataset
        else:
            raise TypeError("Cannot filter dataset, because it is None")



from abc import ABCMeta, abstractmethod


class Io(metaclass=ABCMeta):
    def __init__(self, input_name: str = None, output_name: str = None):
        """
        I/O class to handle files
        :param input_name: Input file name
        :param output_name: Output file name
        """
        self.input_name = input_name
        self.output_name = output_name

    @abstractmethod
    def read(self):
        return

    @abstractmethod
    def write(self):
        return



from .antenna import *  # noqa
from .baseline import * # noqa
from .dataset import * # noqa
from .field import * # noqa
from .ms import * # noqa
from .polarization import * # noqa
from .spectral_window import * # noqa
from .visibility import * # noqa



import xarray as xr
from astropy.units import Quantity
from typing import List, Union
from abc import ABCMeta, abstractmethod
from ..units.units_functions import check_units
import astropy.units as un


class Parameter(metaclass=ABCMeta):
    def __init__(self, data: xr.DataArray = None, cellsize: Union[List[Quantity], Quantity] = None, name: str = None,
                 noise: Quantity = None, chunks: tuple = None, data_gpu: bool = None):
        """
        Class that represents a parameter. This parameter can be passed to Optimizers in order to reconstruct them.
        :param data: Array with the data
        :param cellsize: Cell-size space between data points.
        :param name: Name of the parameter
        :param noise: Noise of the parameter
        :param chunks: Change the chunks of the dask array
        :param data_gpu: Whether to convert data to a cupy array.
        """
        self.data = data
        self.name = name
        self.noise = noise
        self.data_gpu = data_gpu

        if data is not None and chunks is not None:
            self.data = self.data.chunk(chunks)

        if cellsize is not None:
            if isinstance(cellsize, Quantity):
                if cellsize.size == 1:
                    self.cellsize = [cellsize.value, cellsize.value] * cellsize.unit
                else:
                    self.cellsize = cellsize
            else:
                same_units = all(element.unit == cellsize[0].unit for element in cellsize)
                if same_units:
                    self.cellsize = [element.value for element in cellsize] * cellsize[0].unit
                else:
                    raise ValueError("Elements don't have the same units")
        else:
            self.cellsize = cellsize

    @property
    def cellsize(self):
        return self.__cellsize

    @cellsize.setter
    def cellsize(self, val):
        if val is not None:
            if isinstance(val, Quantity):
                if val.size == 1:
                    self.__cellsize = [val.value, val.value] * val.unit
                else:
                    self.__cellsize = val
            else:
                same_units = all(element.unit == val[0].unit for element in val)
                if same_units:
                    self.__cellsize = [element.value for element in val] * val[0].unit
                else:
                    raise ValueError("Elements don't have the same units")
            area = self.__cellsize[0] * self.__cellsize[1]
            if check_units(area, un.steradian):
                self.__pixel_area = area.to(un.steradian)
        else:
            self.__cellsize = val

    @property
    def pixel_area(self):
        return self.__pixel_area

    @abstractmethod
    def calculate_noise(self):
        pass



from .transformer import Transformer
from abc import ABC, ABCMeta, abstractmethod
import xarray as xr
import astropy.units as un
import dask.array as da


class HermitianSymmetry(Transformer, metaclass=ABCMeta):
    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs
               Transformer object arguments
        """
        super().__init__(**kwargs)

    def transform(self):
        """
        Applies hermitian symmetry to the dataset
        """
        for ms in self.input_data.ms_list:
            uvw_aux = xr.apply_ufunc(lambda x: x.value, ms.visibilities.uvw, dask="parallelized",
                                     output_dtypes=[ms.visibilities.uvw.dtype])
            # This is faster than using the indexes as dask arrays
            idx = da.argwhere(uvw_aux[:, 0] < 0).squeeze().compute()
            if len(idx) > 0:
                uvw_aux[idx] *= -1
                ms.visibilities.uvw = uvw_aux * un.m
                ms.visibilities.data[idx] = da.conj(ms.visibilities.data[idx])
                ms.visibilities.model[idx] = da.conj(ms.visibilities.data[idx])



from .gridder import Gridder
from .gridder import bincount, complex_bincount, next_power_of_2, calculate_pix
from abc import ABC
import dask.array as da
import numpy as np
import xarray as xr
from typing import List, Union
import astropy.units as un
from more_itertools import locate
from ..units.lambda_units import lambdas_equivalencies
from ..units import array_unit_conversion
from ..reconstruction import Image
from astropy.units import Quantity


class DirtyMapper(Gridder, ABC):
    def __init__(self, stokes: Union[List[str], str] = None, **kwargs):
        super().__init__(**kwargs)
        if stokes is None:
            self.stokes = ["I", "Q", "U", "V"]
        else:
            if isinstance(stokes, str):
                self.stokes = stokes.split(',')

        if self.imsize is None:
            self.uvgridded_visibilities = None
            self.uvgridded_weights = None
        else:
            self.uvgridded_visibilities = np.zeros((len(self.stokes), self.imsize[0], self.imsize[1]),
                                                   dtype=np.complex64)
            self.uvgridded_weights = np.zeros((len(self.stokes), self.imsize[0], self.imsize[1]), dtype=np.float32)

    def transform(self) -> tuple[Image, Image]:
        idx_I = list(locate(self.stokes, lambda x: x == "I"))
        idx_Q = list(locate(self.stokes, lambda x: x == "Q"))
        idx_U = list(locate(self.stokes, lambda x: x == "U"))
        idx_V = list(locate(self.stokes, lambda x: x == "V"))

        for ms in self.input_data.ms_list:
            chans = ms.spectral_window.chans.compute()
            nchans = ms.spectral_window.nchans
            uvw = ms.visibilities.uvw
            weight = ms.visibilities.weight
            data = ms.visibilities.data
            flag = ms.visibilities.flag
            ncorrs = ms.polarization.ncorrs
            corr_names = ms.polarization.corrs_names
            feed = ms.polarization.feed_kind

            uvw_broadcast = da.tile(uvw, nchans).reshape((len(uvw), nchans, 3))
            chans_broadcast = chans[np.newaxis, :, np.newaxis]

            # uvw_lambdas = da.map_blocks(
            #    lambda x: x.to(un.lambdas, equivalencies=lambdas_equivalencies(restfreq=chans_broadcast)).value,
            #    uvw_broadcast,
            #    dtype=np.float64)
            uvw_lambdas = array_unit_conversion(array=uvw_broadcast, unit=un.lambdas,
                                                equivalencies=lambdas_equivalencies(restfreq=chans_broadcast))

            weight_broadcast = da.repeat(weight.data, nchans, axis=0)
            flag = flag.data.reshape((len(flag.data) * nchans, ncorrs))
            data = data.data.reshape((len(data.data) * nchans, ncorrs))
            weight_broadcast[flag] = 0.0

            idx_corrs = [key for key, value in corr_names.items()]
            idx = calculate_pix(uvw_lambdas, self.uvcellsize, self.imsize)

            bincount_m = idx.max().compute() + 1

            visibility_data = weight_broadcast * data

            bin_count_weights = da.apply_along_axis(bincount, 0, weight_broadcast, idx,
                                                    shape=(bincount_m,),
                                                    dtype=weight_broadcast.dtype)

            bin_count_visibilities = da.apply_along_axis(complex_bincount, 0, visibility_data,
                                                         idx,
                                                         shape=(bincount_m,),
                                                         dtype=visibility_data.dtype)

            idx = idx.compute()
            i_idx, j_idx = np.unravel_index(idx, (self.imsize[0], self.imsize[1]))

            for i in idx_corrs:
                gridded_data = bin_count_visibilities[idx, i].compute()
                gridded_weights = bin_count_weights[idx, i].compute()

                if feed == "linear":
                    if corr_names[i] == "XX":
                        if idx_I:
                            self.uvgridded_visibilities[idx_I[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_I[0], i_idx, j_idx] += gridded_weights

                        if idx_Q:
                            self.uvgridded_visibilities[idx_Q[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_Q[0], i_idx, j_idx] += gridded_weights

                    elif corr_names[i] == "XY":
                        if idx_U:
                            self.uvgridded_visibilities[idx_U[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_U[0], i_idx, j_idx] += gridded_weights

                        if idx_V:
                            self.uvgridded_visibilities[idx_V[0], i_idx, j_idx] += -1.0j * gridded_data / 2
                            self.uvgridded_weights[idx_V[0], i_idx, j_idx] += gridded_weights
                    elif corr_names[i] == "YX":
                        if idx_U:
                            self.uvgridded_visibilities[idx_U[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_U[0], i_idx, j_idx] += gridded_weights

                        if idx_V:
                            self.uvgridded_visibilities[idx_V[0], i_idx, j_idx] += 1.0j * gridded_data / 2
                            self.uvgridded_weights[idx_V[0], i_idx, j_idx] += gridded_weights
                    elif corr_names[i] == "YY":
                        if idx_I:
                            self.uvgridded_visibilities[idx_I[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_I[0], i_idx, j_idx] += gridded_weights

                        if idx_Q:
                            self.uvgridded_visibilities[idx_Q[0], i_idx, j_idx] += -1.0 * gridded_data / 2
                            self.uvgridded_weights[idx_Q[0], i_idx, j_idx] += gridded_weights
                    else:
                        raise ValueError("The correlation does not correspond to a linear feed")
                else:
                    if corr_names[i] == "LL":
                        idx_I = list(locate(self.stokes, lambda x: x == "I"))
                        idx_V = list(locate(self.stokes, lambda x: x == "V"))

                        if idx_I:
                            self.uvgridded_visibilities[idx_I[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_I[0], i_idx, j_idx] += gridded_weights

                        if idx_V:
                            self.uvgridded_visibilities[idx_V[0], i_idx, j_idx] += -1.0j * gridded_data / 2
                            self.uvgridded_weights[idx_V[0], i_idx, j_idx] += gridded_weights
                    elif corr_names[i] == "LR":
                        if idx_Q:
                            self.uvgridded_visibilities[idx_Q[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_Q[0], i_idx, j_idx] += gridded_weights

                        if idx_U:
                            self.uvgridded_visibilities[idx_U[0], i_idx, j_idx] += 1.0j * gridded_data / 2
                            self.uvgridded_weights[idx_U[0], i_idx, j_idx] += gridded_weights
                    elif corr_names[i] == "RL":
                        if idx_Q:
                            self.uvgridded_visibilities[idx_Q[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_Q[0], i_idx, j_idx] += gridded_weights

                        if idx_U:
                            self.uvgridded_visibilities[idx_U[0], i_idx, j_idx] += 1.0j * gridded_data / 2
                            self.uvgridded_weights[idx_U[0], i_idx, j_idx] += gridded_weights
                    elif corr_names[i] == "RR":
                        if idx_I:
                            self.uvgridded_visibilities[idx_I[0], i_idx, j_idx] += gridded_data / 2
                            self.uvgridded_weights[idx_I[0], i_idx, j_idx] += gridded_weights

                        if idx_V:
                            self.uvgridded_visibilities[idx_V[0], i_idx, j_idx] += 1.0j * gridded_data / 2
                            self.uvgridded_weights[idx_V[0], i_idx, j_idx] += gridded_weights
                    else:
                        raise ValueError("The correlation does not correspond to a circular feed")

        # Normalize gridded visibilities
        idx_stokes, idx_i, idx_j = np.where(self.uvgridded_weights > 0.0)
        self.uvgridded_visibilities[idx_stokes, idx_i, idx_j] /= self.uvgridded_weights[idx_stokes, idx_i, idx_j]
        # Make a IFFT
        dirty_maps = np.zeros((len(self.stokes), self.imsize[0], self.imsize[1]),
                              dtype=np.float32)
        dirty_beams = np.zeros((len(self.stokes), self.imsize[0], self.imsize[1]),
                               dtype=np.float32)
        for idx_stokes in range(0, len(self.stokes)):
            dirty_map = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(self.uvgridded_visibilities[idx_stokes])))
            dirty_beam = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(self.uvgridded_weights[idx_stokes])))
            dirty_maps[idx_stokes] = dirty_map.real
            dirty_beams[idx_stokes] = dirty_beam.real / np.max(dirty_beam.real)
        sky_cellsize = self.uvcellsize.to(un.rad, equivalencies=lambdas_equivalencies()) / self.imsize
        self.uvgridded_visibilities = da.from_array(self.uvgridded_visibilities)
        self.uvgridded_weights = da.from_array(self.uvgridded_weights)
        return Image(data=xr.DataArray(da.from_array(dirty_maps)), cellsize=sky_cellsize), \
               Image(data=xr.DataArray(da.from_array(dirty_beams)), cellsize=sky_cellsize)



from abc import ABC

from ..wscheme import WeightingScheme


class Radial(WeightingScheme, ABC):
    def __init__(self, use_w: bool = False, **kwargs):
        """
        Class that represents the Radial Weighting Scheme.
        :param use_w: Whether to use the w position to calculate the uv-distance
        :param kwargs: WeightingScheme arguments
        """
        super().__init__(**kwargs)
        self.use_w = use_w

    def transform(self) -> None:
        """
        This function calculates the radial weights and transforms them according to this scheme,
        """
        for ms in self.input_data.ms_list:
            dist = ms.visibilities.get_uvw_distance(self.use_w)
            weight = ms.visibilities.weight
            weight *= dist



import numpy as np
from abc import ABC
from typing import List, Union
import astropy.units as un
import xarray as xr
import dask.array as da
from astropy.units import Quantity
from .transformer import Transformer
from ..reconstruction.convolution import CKernel
from ..reconstruction import Image
from ..units import lambdas_equivalencies
from ..base import MS


def bincount(weights, x):
    return np.bincount(x, weights)


def complex_bincount(weights, x):
    real_w = weights.real
    imag_w = weights.imag
    return np.bincount(x, real_w) + 1j * np.bincount(x, imag_w)


def next_power_of_2(x):
    return 1 if x == 0 else 1 << (x - 1).bit_length()


def calculate_pix(uvw, uvcellsize, imsize, ref_freq=None):
    """

    Parameters
    ----------
    uvw : uvw position in Fourier space
    uvcellsize : Cell-size in Fourier space
    imsize : Image size
    ref_freq (Optional) : Reference Frequency

    Returns
    -------
    idx : Pixel 1D index of the cells on which the uv coordinates fall

    """
    m = imsize[0]
    n = imsize[1]

    if ref_freq is not None:
        duv = uvcellsize.to(un.m, equivalencies=lambdas_equivalencies(ref_freq))
    else:
        duv = uvcellsize

    if uvw.ndim == 3:
        uv = uvw[:, :, 0:2] / duv
    else:
        uv = uvw[:, 0:2] / duv

    if isinstance(uv, xr.DataArray):
        uv = xr.apply_ufunc(lambda x: x.value, uv, dask="parallelized", output_dtypes=[uv.dtype]).round().astype(np.int32)
    elif isinstance(uv, da.core.Array):
        uv = da.map_blocks(lambda x: x.value, uv, dtype=uv.dtype).round().astype(np.int32)
    else:
        raise TypeError("The array is not a dask nor xarray type")

    if uvw.ndim == 3:
        uv = uv.reshape((uv.shape[0]*uv.shape[1], uv.shape[2]))

    u_pix = uv[:, 0] + m // 2
    v_pix = uv[:, 1] + n // 2
    idx = n * v_pix + u_pix

    if isinstance(idx, xr.DataArray):
        return idx.data
    else:
        return idx


class Gridder(Transformer, ABC):
    def __init__(self, image: Image = None, imsize: Union[List[int], int] = None, uvcellsize: Quantity = None,
                 ckernel_object: CKernel = None, **kwargs):
        """
        Class that represents the interferometric gridding
        :param imsize: Image size
        :param uvcellsize: Cell size in Fourier-space
        :param ckernel_object: Convolution Kernel
        :param kwargs: Transformer arguments
        """
        super().__init__(**kwargs)
        self.ckernel_object = ckernel_object
        self.image = image
        if image is None:
            if isinstance(imsize, int):
                self.imsize = [imsize, imsize]
            else:
                self.imsize = imsize

            if isinstance(uvcellsize, Quantity):
                if uvcellsize.shape[0] > 1:
                    self.uvcellsize = uvcellsize
                else:
                    self.uvcellsize = Quantity([uvcellsize, uvcellsize, uvcellsize])
            else:
                raise TypeError("uvcellsize must be a Quantity instance")
        else:
            self.imsize = image.imsize
            self.uvcellsize = image.cellsize.to(un.lambdas, equivalencies=lambdas_equivalencies()) / self.imsize

    def grid_weights(self, ms: MS = None):
        if ms is not None:

            ref_freq = ms.spectral_window.ref_frequency
            uvw = ms.visibilities.uvw
            weight = ms.visibilities.weight
            idx = calculate_pix(uvw, self.uvcellsize, self.imsize, ref_freq)
            bincount_m = idx.max().compute() + 1
            bin_count = da.apply_along_axis(bincount, 0, weight.data, idx, shape=(bincount_m,),
                                            dtype=weight.data.dtype)
            w_k = bin_count[idx]  # Gridded weight for each non-gridded u,v coordinate
            return idx, bin_count, w_k

        else:
            raise ValueError("MS cannot be Nonetype")

    # This does the convolutional gridding
    def transform(self) -> None:

        for ms in self.input_data.ms_list:
            ref_freq = ms.spectral_window.ref_frequency
            uvw = ms.visibilities.uvw
            weight = ms.visibilities.weight
            data = ms.visibilities.data
            flag = ms.visibilities.flag

            idx = calculate_pix(uvw, self.uvcellsize, self.imsize, ref_freq)
            ncorrs = ms.polarization.ncorrs
            nchans = ms.spectral_window.nchans
            # chans = ms.spectral_window.chans.compute()
            baselines = da.unique(ms.visibilities.baseline.data).compute()
            weight_list = []
            data_list = []
            for baseline_id in baselines:
                row_id = da.argwhere(ms.visibilities.baseline.data == baseline_id).squeeze().compute()
                idx_per_baseline = idx[row_id]
                weight_per_baseline = weight[row_id]
                flag_per_baseline = flag[row_id]  # (rows, nchans, ncorrs)
                data_per_baseline = data[row_id]
                bincount_m = idx_per_baseline.max().compute() + 1
                # Grid the weights (rows, ncorrs)
                bin_count_weights = da.apply_along_axis(bincount, 0, weight_per_baseline.data, idx_per_baseline,
                                                        shape=(bincount_m,),
                                                        dtype=weight_per_baseline.data.dtype)
                gridded_weights = bin_count_weights[idx_per_baseline]

                # Grid the complex data (rows, nchans, ncorrs) (w * data)
                weight_broadcast = da.tile(weight_per_baseline, nchans).reshape(
                    (len(weight_per_baseline), nchans, ncorrs))
                weight_broadcast[flag_per_baseline] = 0.0
                visibility_data_per_baseline = weight_broadcast * data_per_baseline
                bin_count_visibilities = da.apply_along_axis(complex_bincount, 0, visibility_data_per_baseline.data,
                                                             idx_per_baseline,
                                                             shape=(bincount_m,),
                                                             dtype=visibility_data_per_baseline.data.dtype)

                bin_count_weight_broadcast = da.apply_along_axis(bincount, 0, weight_broadcast,
                                                                 idx_per_baseline,
                                                                 shape=(bincount_m,),
                                                                 dtype=weight_broadcast.dtype)
                gridded_data = bin_count_visibilities[idx_per_baseline] / bin_count_weight_broadcast[idx_per_baseline]
                weight_list.append(gridded_weights)
                data_list.append(gridded_data)
            print(data_list)
            break



from abc import ABC
from ..wscheme import WeightingScheme


class Uniform(WeightingScheme, ABC):
    def __init__(self, **kwargs):
        """
        :param kwargs: WeightingScheme arguments
        """
        super().__init__(**kwargs)

    def transform(self) -> None:
        """
        This function calculates the uniform weights and transforms them according to this scheme,
        """
        if self.gridder is not None:
            dataset = self.input_data
            for ms in dataset.ms_list:
                weight = ms.visibilities.weight
                idx, bin_count, w_k = self.gridder.grid_weights(ms)
                weight /= w_k
        else:
            raise TypeError("gridding attribute cannot be None when calculting uniform weights")



import numpy as np
import astropy.units as u
import dask.array as da
import xarray as xr
from astropy.constants import c
from .antenna import Antenna
from .baseline import Baseline
from .ms import MS
from ..reconstruction import PSF
from ..units.lambda_units import lambdas_equivalencies
from astropy.units import Quantity
from typing import List
import logging


def calc_beam_size(s_uu, s_vv, s_uv) -> tuple:
    """

    Parameters
    ----------
    s_uu : float
          Weighted sum of u^2.
    s_vv : float
          Weighted sum of v^2.
    s_uv : float
          Weighted sum of u*v.

    Returns
    -------
    tuple
        Beam major, minor and position angle in radians.
    """
    uv_squared = s_uv ** 2
    u_minus_v = s_uu - s_vv
    u_plus_v = s_uu + s_vv
    sqrt_par = np.sqrt(u_minus_v ** 2 + 4.0 * uv_squared)
    bmaj = 2.0 * np.sqrt(np.log(2.0)) / np.pi / np.sqrt(u_plus_v - sqrt_par)  # Major axis in radians
    bmin = 2.0 * np.sqrt(np.log(2.0)) / np.pi / np.sqrt(u_plus_v + sqrt_par)  # Minor axis in radians
    bpa = -0.5 * np.arctan2(2.0 * s_uv, u_minus_v)  # Angle in radians
    return bmaj * u.rad, bmin * u.rad, bpa * u.rad


class Dataset:
    def __init__(self, antenna: Antenna = None, baseline: Baseline = None,
                 spectral_window_dataset: List[xr.Dataset] = None, ms_list: List[MS] = None, psf_parhands: PSF = None,
                 psf_crossedhands: PSF = None, feed_kind: str = None):
        """

        Parameters
        ----------
        antenna : Full Antenna object
        baseline : Full Baseline object
        spectral_window_dataset : Full Spectral Window dataset
        ms_list : List of separated MS
        psf_parhands : PSF object for parallel hands
        psf_crossedhands : PSF object for crossed hands
        feed_kind : Kind of feed of the dataset. E.g "linear", "circular".
        """

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self.antenna = antenna
        self.baseline = baseline
        self.spectral_window_dataset = spectral_window_dataset
        self.ms_list = ms_list
        self.psf_parhands = psf_parhands
        self.psf_crossedhands = psf_crossedhands
        self.feed_kind = feed_kind

        self.max_nu = 0.0 * u.Hz  # Maximum frequency in Hz
        self.min_nu = 0.0 * u.Hz  # Minimum frequency in Hz
        self.ref_nu = 0.0 * u.Hz  # Reference frequency in Hz

        self.lambda_min = 0.0 * u.m  # Minimum wavelength in meters
        self.lambda_max = 0.0 * u.m  # Maximum wavelength in meters
        self.lambda_ref = 0.0 * u.m  # Reference wavelength in meters

        self.max_baseline = 0.0 * u.m  # Maximum baseline in meters
        self.min_baseline = 0.0 * u.m  # Minimum baseline in meters

        self.max_antenna_diameter = 0.0 * u.m  # Maximum antenna diameter in meters
        self.min_antenna_diameter = 0.0 * u.m  # Minimum antenna diameter in meters

        self.theo_resolution = 0.0 * u.rad  # Theoretical resolution in radians
        self.fov = 0.0 * u.rad  # Field-of-view in radians

        self.ndatasets = 0

        self.corr_weight_sum = None
        self.parhands_weight_sum = 0.0
        self.crossedhands_weight_sum = 0.0
        self.weights_sum = 0.0

        if spectral_window_dataset is not None:
            max_freqs = []
            min_freqs = []
            ref_freqs = []
            for spw in spectral_window_dataset:
                max_freqs.append(spw.CHAN_FREQ.max().data)
                min_freqs.append(spw.CHAN_FREQ.min().data)
                ref_freqs.append(spw.REF_FREQUENCY.data)

            max_freqs = list(da.compute(*max_freqs))
            min_freqs = list(da.compute(*min_freqs))

            self.max_nu = max(max_freqs) * u.Hz
            self.min_nu = min(min_freqs) * u.Hz
            self.ref_nu = np.median(np.array(ref_freqs)) * u.Hz

            self.lambda_min = c / self.max_nu
            self.lambda_min = self.lambda_min.to(u.m)
            self.lambda_max = c / self.min_nu
            self.lambda_max = self.lambda_max.to(u.m)
            self.lambda_ref = c / self.ref_nu
            self.lambda_ref = self.lambda_ref.to(u.m)

        if antenna is not None:
            self.max_antenna_diameter = antenna.max_diameter
            self.min_antenna_diameter = antenna.min_diameter

        if baseline is not None:
            self.max_baseline = baseline.max_baseline
            self.min_baseline = baseline.min_baseline

        if antenna is not None and baseline is not None and spectral_window_dataset is not None:
            self.theo_resolution = (self.lambda_min / self.max_baseline) * u.rad
            self.fov = (self.lambda_max / self.max_antenna_diameter) * u.rad

        if ms_list is not None:
            self.ndatasets = len(self.ms_list)
            self.check_feed()

            if self.psf_parhands is None and self.psf_crossedhands is None:
                self.calculate_PSF()

    def check_feed(self) -> None:
        """
        Function to check if a feed of the dataset is
        linear, circular or mixed
        """
        feed_list = []
        for ms in self.ms_list:
            feed_list.append(ms.polarization.feed_kind)
        if all(x == "linear" for x in feed_list):
            self.feed_kind = "linear"
        elif all(x == "circular" for x in feed_list):
            self.feed_kind = "circular"
        else:
            self.feed_kind = "mixed"

    def max_ncorrs(self) -> int:
        ncorrs = []
        for ms in self.ms_list:
            ncorrs.append(ms.polarization.ncorrs)
        return max(ncorrs)

    def calculate_weights_sum(self) -> None:

        if self.feed_kind == "linear":
            hands_dict = {'XX': [], 'YX': [], 'XY': [], 'YY': []}
        elif self.feed_kind == "circular":
            hands_dict = {'LL': [], 'RL': [], 'LR': [], 'RR': []}
        else:
            hands_dict = {'XX': [], 'YX': [], 'XY': [], 'YY': [], 'LL': [], 'RL': [], 'LR': [], 'RR': []}

        for ms in self.ms_list:
            weight = ms.visibilities.weight
            ncorrs = ms.polarization.ncorrs
            corr_names = ms.polarization.corrs_names

            for idx_corr in range(0, ncorrs):
                hands_dict[corr_names[idx_corr]].append(da.sum(weight[:, idx_corr]))

        hands_dict = da.compute(hands_dict)[0]
        for i in hands_dict:
            hands_dict[i] = sum(hands_dict[i])

        if self.feed_kind == "linear":
            self.parhands_weight_sum = hands_dict["XX"] + hands_dict["YY"]
            self.crossedhands_weight_sum = hands_dict["YX"] + hands_dict["XY"]
        elif self.feed_kind == "circular":
            self.parhands_weight_sum = hands_dict["LL"] + hands_dict["RR"]
            self.crossedhands_weight_sum = hands_dict["LR"] + hands_dict["RL"]
        else:
            self.parhands_weight_sum = hands_dict["XX"] + hands_dict["YY"] + hands_dict["LL"] + hands_dict["RR"]
            self.crossedhands_weight_sum = hands_dict["YX"] + hands_dict["XY"] + hands_dict["LR"] + hands_dict["RL"]

        self.weights_sum = self.parhands_weight_sum + self.crossedhands_weight_sum
        self.corr_weight_sum = hands_dict

    def calculate_PSF(self) -> None: # TODO: Calculate PSF for all stokes if parameter is None, or for a stokes
        # parameter input
        """
        Function that calculates the PSF properties (bmaj, bmin and bpa) analytically using
        (u,v) positions and the weights
        """
        self.calculate_weights_sum()
        s_uu = [[], []]
        s_vv = [[], []]
        s_uv = [[], []]
        for ms in self.ms_list:
            chans = ms.spectral_window.chans.compute()
            nchans = ms.spectral_window.nchans
            uvw = ms.visibilities.uvw
            weight = ms.visibilities.weight
            flag = ms.visibilities.flag
            ncorrs = ms.polarization.ncorrs
            corr_names = ms.polarization.corrs_names

            uvw_broadcast = da.tile(uvw, nchans * ncorrs).reshape((len(uvw), nchans, ncorrs, 3))
            chans_broadcast = chans[np.newaxis, :, np.newaxis, np.newaxis]

            uvw_lambdas = da.map_blocks(
                lambda x: x.to(u.lambdas, equivalencies=lambdas_equivalencies(restfreq=chans_broadcast)).value,
                uvw_broadcast,
                dtype=np.float64)

            weight_broadcast = da.tile(weight, nchans).reshape((len(weight), nchans, ncorrs))
            weight_broadcast[flag.data] = 0.0

            idx_corr_linear = [key for key, value in corr_names.items()
                               if value == "XX" or value == "YY" or value == "LL" or value == "RR"]
            idx_corr_crossed = [key for key, value in corr_names.items()
                                if value == "XY" or value == "YX" or value == "LR" or value == "RL"]

            if idx_corr_linear:
                u_lambdas = uvw_lambdas[:, :, idx_corr_linear, 0]
                v_lambdas = uvw_lambdas[:, :, idx_corr_linear, 1]
                s_uu[0].append(da.sum((u_lambdas ** 2) * weight_broadcast[:, :, idx_corr_linear]))
                s_vv[0].append(da.sum((v_lambdas ** 2) * weight_broadcast[:, :, idx_corr_linear]))
                s_uv[0].append(da.sum((u_lambdas * v_lambdas) * weight_broadcast[:, :, idx_corr_linear]))

            if idx_corr_crossed:
                u_lambdas = uvw_lambdas[:, :, idx_corr_crossed, 0]
                v_lambdas = uvw_lambdas[:, :, idx_corr_crossed, 1]
                s_uu[1].append(da.sum((u_lambdas ** 2) * weight_broadcast[:, :, idx_corr_crossed]))
                s_vv[1].append(da.sum((v_lambdas ** 2) * weight_broadcast[:, :, idx_corr_crossed]))
                s_uv[1].append(da.sum((u_lambdas * v_lambdas) * weight_broadcast[:, :, idx_corr_crossed]))

        s_uu_parhands = np.asarray(da.compute(*s_uu[0])).sum()
        s_vv_parhands = np.asarray(da.compute(*s_vv[0])).sum()
        s_uv_parhands = np.asarray(da.compute(*s_uv[0])).sum()

        s_uu_crossedhands = np.asarray(da.compute(*s_uu[1])).sum()
        s_vv_crossedhands = np.asarray(da.compute(*s_vv[1])).sum()
        s_uv_crossedhands = np.asarray(da.compute(*s_uv[1])).sum()

        if self.parhands_weight_sum > 0.0:
            s_uu_parhands /= self.parhands_weight_sum
            s_vv_parhands /= self.parhands_weight_sum
            s_uv_parhands /= self.parhands_weight_sum
            bmaj_par, bmin_par, bpa_par = calc_beam_size(s_uu_parhands, s_vv_parhands, s_uv_parhands)
            self.psf_parhands = PSF(sigma=Quantity([bmaj_par, bmin_par]), theta=bpa_par)

        if self.crossedhands_weight_sum > 0.0:
            s_uu_crossedhands /= self.crossedhands_weight_sum
            s_vv_crossedhands /= self.crossedhands_weight_sum
            s_uv_crossedhands /= self.crossedhands_weight_sum
            bmaj_crossed, bmin_crossed, bpa_crossed = calc_beam_size(s_uu_crossedhands, s_vv_crossedhands,
                                                                     s_uv_crossedhands)
            self.psf_crossedhands = PSF(sigma=Quantity([bmaj_crossed, bmin_crossed]), theta=bpa_crossed)



import logging
import dask.array as da
import astropy.units as un
from ..units import lambdas_equivalencies
from ..units import xarray_unit_conversion
import xarray as xr


class Visibility:
    def __init__(self, dataset: xr.Dataset = None):
        """
        Class that represents the main table visibilities on the partitioned Measurement Set.

        Parameters
        ----------
        dataset : xarray.Dataset
                  xarray Dataset of the main table on the partitioned Measurement Set
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self.nrows = 0

        self.dataset = dataset
        if dataset is not None:
            self.nrows = len(self.dataset.ROWID)
            self.rows = self.dataset.ROWID
            self.uvw = self.dataset.UVW * un.m
            self.weight = self.dataset.WEIGHT
            self.flag = self.dataset.FLAG
            self.data = self.dataset.DATA
            self.model = xr.zeros_like(self.data)
            self.antenna1 = self.dataset.ANTENNA1
            self.antenna2 = self.dataset.ANTENNA2
            self.baseline = self.dataset.BASELINE
            self.unflagged_rows, self.unflagged_chans, self.unflagged_corrs = da.nonzero(~self.flag.data)

    def get_column(self, column=None, as_xarray=True):
        if column is not None:
            if as_xarray:
                return self.dataset[column]
            else:
                return self.dataset[column].data.compute()

    def get_uvw_lambda(self, nu=None, as_xarray: bool = True):
        if nu is None:
            raise ValueError("Error: Input frequency is not present")

        if not as_xarray:
            return self.uvw.data.compute().to(un.lambdas, equivalencies=lambdas_equivalencies(restfreq=nu))
        else:
            # runs function decompose on a lazy manner over all items of the dask array
            uvw = xarray_unit_conversion(self.uvw, un.lambdas, lambdas_equivalencies(restfreq=nu))

            return uvw

    def get_uvw_distance(self, count_w: bool = True):
        u = self.uvw[:, 0]
        v = self.uvw[:, 1]
        if count_w:
            w = self.uvw[:, 2]
        else:
            w = 0
        distance = xr.ufuncs.sqrt(u ** 2 + v ** 2 + w ** 2)
        return distance



import xarray as xr
import logging
import astropy.units as u


class Antenna:
    def __init__(self, dataset: xr.Dataset = None):
        """
        Class that represents the antenna MS table of an interferometer

        Parameters
        ----------
        dataset : xarray.Dataset
                  Full antenna table xarray dataset.
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self.dataset = dataset

        self.max_diameter = 0.0 * u.m
        self.min_diameter = 0.0 * u.m
        if dataset is not None:
            self.max_diameter = self.dataset.DISH_DIAMETER.data.max().compute() * u.m
            self.min_diameter = self.dataset.DISH_DIAMETER.data.min().compute() * u.m
